@@ -28,6 +28,7 @@
 #include <kio/netaccess.h>
 #include <krandomsequence.h>
 #include <kfilterdev.h>
+#include <kautosavefile.h>
 
 #include "keduvocexpression.h"
 #include "keduvoclesson.h"
@@ -35,13 +36,9 @@
 #include "keduvocwordtype.h"
 #include "keduvockvtmlwriter.h"
 #include "keduvockvtml2writer.h"
-#include "keduvoccsvreader.h"
 #include "keduvoccsvwriter.h"
-#include "keduvockvtml2reader.h"
-#include "keduvocwqlreader.h"
-#include "keduvocpaukerreader.h"
-#include "keduvocvokabelnreader.h"
-#include "keduvocxdxfreader.h"
+#include "readerwriters/readermanager.h"
+#include "readerwriters/readerbase.h"
 
 #define WQL_IDENT      "WordQuiz"
 
@@ -59,6 +56,7 @@ public:
         m_lessonContainer = 0;
         m_wordTypeContainer = 0;
         m_leitnerContainer = 0;
+        m_autosave = new KAutoSaveFile;
         init();
     }
 
@@ -68,8 +66,13 @@ public:
 
     KEduVocDocument* q;
 
+    /** autosave file used to provide locking access to the underlying file
+     * Note: It is a pointer to allow locking a new file, saving results and
+     * then transfering the lock to m_autosave without risking loss of lock.
+     * See saveAs for clarification*/
+    KAutoSaveFile            *m_autosave;
+
     bool                      m_dirty;
-    KUrl                      m_url;
 
     // save these to document
     QList<KEduVocIdentifier>  m_identifiers;
@@ -100,6 +103,14 @@ public:
     KEduVocLesson * m_lessonContainer;
     KEduVocWordType * m_wordTypeContainer;
     KEduVocLeitnerBox * m_leitnerContainer;
+
+    /** Creates an autosave file for fpath in order to lock the file
+     *   @param autosave a reference to an allocated KAutosave
+     *   @param fpath File path to the file to be locked
+     *   @param flags Describes how to deal with locked file etc.
+     *   @return ErrorCode where NoError is success
+     *   */
+    KEduVocDocument::ErrorCode initializeKAutoSave(KAutoSaveFile &autosave,  QString const &fpath,  FileHandlingFlags flags) const;
 };
 
 KEduVocDocument::KEduVocDocumentPrivate::~KEduVocDocumentPrivate()
@@ -107,6 +118,9 @@ KEduVocDocument::KEduVocDocumentPrivate::~KEduVocDocumentPrivate()
     delete m_lessonContainer;
     delete m_wordTypeContainer;
     delete m_leitnerContainer;
+
+    m_autosave->releaseLock();
+    delete m_autosave;
 }
 
 void KEduVocDocument::KEduVocDocumentPrivate::init()
@@ -127,7 +141,7 @@ void KEduVocDocument::KEduVocDocumentPrivate::init()
     m_dirty = false;
     m_queryorg = "";
     m_querytrans = "";
-    m_url.setFileName( i18n( "Untitled" ) );
+    m_autosave->setManagedFile( i18n( "Untitled" ) );
     m_author = "";
     m_title = "";
     m_comment = "";
@@ -140,10 +154,35 @@ void KEduVocDocument::KEduVocDocumentPrivate::init()
 }
 
 
+KEduVocDocument::ErrorCode KEduVocDocument::KEduVocDocumentPrivate::initializeKAutoSave(KAutoSaveFile &autosave,  QString const & fpath,  FileHandlingFlags flags) const {
+
+    QList<KAutoSaveFile *> staleFiles = KAutoSaveFile::staleFiles( fpath );
+    if ( !staleFiles.isEmpty()) {
+        if ( flags & FileIgnoreLock ) {
+            foreach( KAutoSaveFile *f,  staleFiles ) {
+                f->open( QIODevice::ReadWrite );
+                f->remove();
+                delete f;
+            }
+        } else {
+            kWarning() << i18n( "Cannot lock file %1", fpath );
+            return FileLocked;
+        }
+    }
+
+    autosave.setManagedFile( fpath );
+    if ( !autosave.open( QIODevice::ReadWrite ) ) {
+        kWarning() << i18n( "Cannot lock file %1", autosave.fileName() );
+        return FileCannotLock;
+    }
+
+    return NoError;
+}
+
+
 KEduVocDocument::KEduVocDocument( QObject *parent )
         : QObject( parent ), d( new KEduVocDocumentPrivate( this ) )
 {
-    kDebug() << "constructor done";
 }
 
 
@@ -236,130 +275,81 @@ KEduVocDocument::FileType KEduVocDocument::detectFileType( const QString &fileNa
 }
 
 
-int KEduVocDocument::open( const KUrl& url )
+/// @todo When the API major version number increments remove this function
+int KEduVocDocument::open( const KUrl& url ) {
+    return open( url, FileDefaultHandling );
+}
+
+KEduVocDocument::ErrorCode KEduVocDocument::open( const KUrl& url,  FileHandlingFlags flags)
 {
     // save csv delimiter to preserve it in case this is a csv document
     QString csv = d->m_csvDelimiter;
     // clear all other properties
     d->init();
     if ( !url.isEmpty() ) {
-        d->m_url = url;
+        setUrl( url );
     }
     d->m_csvDelimiter = csv;
 
-    bool read = false;
+    KEduVocDocument::ErrorCode errStatus = Unknown;
     QString errorMessage = i18n( "<qt>Cannot open file<br /><b>%1</b></qt>", url.path() );
     QString temporaryFile;
-    if ( KIO::NetAccess::download( url, temporaryFile, 0 ) ) {
-        QIODevice * f = KFilterDev::deviceForFile( temporaryFile );
-
-        if ( !f->open( QIODevice::ReadOnly ) ) {
-            kError() << errorMessage;
-            delete f;
-            return FileCannotRead;
-        }
-
-        FileType ft = detectFileType( temporaryFile );
-
-        switch ( ft ) {
-            case Kvtml: {
-                kDebug(1100) << "Reading KVTML document...";
-                KEduVocKvtml2Reader kvtmlReader( f );
-                read = kvtmlReader.readDoc( this );
-                if ( !read ) {
-                    errorMessage = kvtmlReader.errorMessage();
-                }
-            }
-            break;
-
-            case Wql: {
-                kDebug(1100) << "Reading WordQuiz (WQL) document...";
-                KEduVocWqlReader wqlReader( f );
-                d->m_url.setFileName( i18n( "Untitled" ) );
-                read = wqlReader.readDoc( this );
-                if ( !read ) {
-                    errorMessage = wqlReader.errorMessage();
-                }
-            }
-            break;
-
-            case Pauker: {
-                kDebug(1100) << "Reading Pauker document...";
-                KEduVocPaukerReader paukerReader( this );
-                d->m_url.setFileName( i18n( "Untitled" ) );
-                read = paukerReader.read( f );
-                if ( !read ) {
-                    errorMessage = i18n( "Parse error at line %1, column %2:\n%3", paukerReader.lineNumber(), paukerReader.columnNumber(), paukerReader.errorString() );
-                }
-            }
-            break;
-
-            case Vokabeln: {
-                kDebug(1100) << "Reading Vokabeln document...";
-                KEduVocVokabelnReader vokabelnReader( f );
-                d->m_url.setFileName( i18n( "Untitled" ) );
-                read = vokabelnReader.readDoc( this );
-                if ( !read ) {
-                    errorMessage = vokabelnReader.errorMessage();
-                }
-            }
-            break;
-
-            case Csv: {
-                kDebug(1100) << "Reading CVS document...";
-                KEduVocCsvReader csvReader( f );
-                read = csvReader.readDoc( this );
-                if ( !read ) {
-                    errorMessage = csvReader.errorMessage();
-                }
-            }
-            break;
-
-            case Xdxf: {
-                kDebug(1100) << "Reading XDXF document...";
-                KEduVocXdxfReader xdxfReader( this );
-                d->m_url.setFileName( i18n( "Untitled" ) );
-                read = xdxfReader.read( f );
-                if ( !read ) {
-                    errorMessage = i18n( "Parse error at line %1, column %2:\n%3", xdxfReader.lineNumber(), xdxfReader.columnNumber(), xdxfReader.errorString() );
-                }
-            }
-            break;
-
-            default: {
-                kDebug(1100) << "Reading KVTML document (fallback)...";
-                KEduVocKvtml2Reader kvtmlReader( f );
-                read = kvtmlReader.readDoc( this );
-                if ( !read ) {
-                    errorMessage = kvtmlReader.errorMessage();
-                }
-            }
-        }
-
-        if ( !read ) {
-            QString msg = i18n( "Could not open or properly read \"%1\"\n(Error reported: %2)", url.path(), errorMessage );
-            kError() << msg << i18n( "Error Opening File" );
-            ///@todo make the readers return int, pass on the error message properly
-            delete f;
-            return FileReaderFailed;
-        }
-
-        f->close();
-        delete f;
-        KIO::NetAccess::removeTempFile( temporaryFile );
+    if ( ! KIO::NetAccess::download( url, temporaryFile, 0 ) ) {
+        return FileDoesNotExist;
     }
 
-    if ( !read ) {
-        return FileReaderFailed;
+    ErrorCode autosaveError = d->initializeKAutoSave( *d->m_autosave,  temporaryFile, flags );
+    if ( autosaveError != NoError) {
+        return autosaveError;
     }
 
-    setModified(false);
-    return 0;
+    QIODevice * f = KFilterDev::deviceForFile( temporaryFile );
+    if ( f->open( QIODevice::ReadOnly ) ) {
+
+        ReaderManager::ReaderPtr reader( ReaderManager::reader( *f ) );
+        errStatus = reader->read( *this );
+
+        if ( errStatus != KEduVocDocument::NoError ) {
+            errorMessage = i18n( "Could not open or properly read \"%1\"\n(Error reported: %2)"
+                                , url.path(), reader->errorMessage() );
+        }
+    } else {
+        errStatus = FileCannotRead;
+    }
+
+    f->close();
+    delete f;
+    KIO::NetAccess::removeTempFile( temporaryFile );
+
+    if ( errStatus == KEduVocDocument::NoError ) {
+        setModified(false);
+    } else {
+        kWarning() << errorMessage;
+    }
+
+    return errStatus;
 }
 
+void KEduVocDocument::close() {
+    d->m_autosave->releaseLock();
+}
 
+/// @todo When the API major version number increments remove this function
 int KEduVocDocument::saveAs( const KUrl & url, FileType ft, const QString & generator )
 {
+    const QString oldgenerator( d->m_generator );
+    d->m_generator = generator;
+
+    ErrorCode err = saveAs( url,  ft,  FileDefaultHandling );
+
+    d->m_generator = oldgenerator;
+
+    return err;
+}
+
+KEduVocDocument::ErrorCode KEduVocDocument::saveAs( const KUrl & url, FileType ft, FileHandlingFlags flags)
+{
+
     KUrl tmp( url );
 
     if ( ft == Automatic ) {
@@ -372,34 +362,57 @@ int KEduVocDocument::saveAs( const KUrl & url, FileType ft, const QString & gene
         }
     }
 
-    QFile f( tmp.path() );
+    QString errorMessage = i18n( "Cannot write to file %1", tmp.path() );
 
+    KAutoSaveFile *autosave;
+
+    //If we don't care about the lock always create a new one
+    //If we are changing files create a new lock
+    if ( ( flags & FileIgnoreLock )
+        || ( d->m_autosave->managedFile() != tmp.path() ) ) {
+
+        autosave = new KAutoSaveFile;
+        ErrorCode autosaveError = d->initializeKAutoSave( *autosave,  tmp.path(), flags );
+        if ( autosaveError != NoError) {
+            delete autosave;
+            return autosaveError;
+        }
+    }
+
+    //We care about the lock and we think we still have it.
+    else {
+         autosave = d->m_autosave;
+        //Is the lock still good?
+        if ( !autosave->exists( ) ) {
+            return FileCannotLock;
+        }
+    }
+
+    QFile f( tmp.path() );
     if ( !f.open( QIODevice::WriteOnly ) ) {
-        kError() << i18n( "Cannot write to file %1", tmp.path() );
+        kError() << errorMessage;
         return FileCannotWrite;
     }
 
-    KUrl oldUrl = d->m_url;
     bool saved = false;
-    d->m_url = tmp;
 
     switch ( ft ) {
         case Kvtml: {
             // write version 2 file
             KEduVocKvtml2Writer kvtmlWriter( &f );
-            saved = kvtmlWriter.writeDoc( this, generator );
+            saved = kvtmlWriter.writeDoc( this, d->m_generator );
         }
         break;
         ///@todo port me
 //         case Kvtml1: {
 //             // write old version 1 file
 //             KEduVocKvtmlWriter kvtmlWriter( &f );
-//             saved = kvtmlWriter.writeDoc( this, generator );
+//             saved = kvtmlWriter.writeDoc( this, d->m_generator );
 //         }
 //         break;
         case Csv: {
             KEduVocCsvWriter csvWriter( &f );
-            saved = csvWriter.writeDoc( this, generator );
+            saved = csvWriter.writeDoc( this, d->m_generator );
         }
         break;
         default: {
@@ -411,13 +424,22 @@ int KEduVocDocument::saveAs( const KUrl & url, FileType ft, const QString & gene
     f.close();
 
     if ( !saved ) {
+        if ( autosave != d->m_autosave ) {
+            delete autosave;
+        }
         kError() << "Error Saving File" << tmp.path();
-        d->m_url = oldUrl;
         return FileWriterFailed;
     }
 
+    if ( autosave != d->m_autosave ) {
+        //The order is important: release old lock, delete old locker and then claim new locker.
+        d->m_autosave->releaseLock();
+        delete d->m_autosave;
+        d->m_autosave = autosave;
+    }
+
     setModified( false );
-    return 0;
+    return NoError;
 }
 
 QByteArray KEduVocDocument::toByteArray(const QString &generator)
@@ -730,18 +752,18 @@ KEduVocLeitnerBox * KEduVocDocument::leitnerContainer()
 
 KUrl KEduVocDocument::url() const
 {
-    return d->m_url;
+    return d->m_autosave->managedFile();
 }
 
 void KEduVocDocument::setUrl( const KUrl& url )
 {
-    d->m_url = url;
+    d->m_autosave->setManagedFile(url);
 }
 
 QString KEduVocDocument::title() const
 {
     if ( d->m_title.isEmpty() )
-        return d->m_url.fileName();
+        return d->m_autosave->managedFile().fileName();
     else
         return d->m_title;
 }
@@ -749,6 +771,7 @@ QString KEduVocDocument::title() const
 void KEduVocDocument::setTitle( const QString & title )
 {
     d->m_title = title;
+    ///@todo decouple document title and root lesson title
     d->m_lessonContainer->setName(title);
     setModified(true);
 }
@@ -909,6 +932,10 @@ QString KEduVocDocument::errorDescription( int errorCode )
         return i18n("The file reader failed.");
     case FileDoesNotExist:
         return i18n("The file does not exist.");
+    case FileLocked:
+        return i18n("The file is locked by another process.");
+    case FileCannotLock:
+        return i18n("The lock file can't be created.");
     case Unknown:
     default:
         return i18n("Unknown error.");
@@ -916,4 +943,3 @@ QString KEduVocDocument::errorDescription( int errorCode )
 }
 
 #include "keduvocdocument.moc"
-
